"""This module provides a function which reads in the data,
performs the guinier fit with a given algotithm and reates the input."""

__authors__ = ["Martha Brennich"]
__contact__ = "martha.brennich@googlemail.com"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "11/09/2024"
__status__ = "development"
__docformat__ = "restructuredtext"

import sys
import logging
import platform
import os
from pathlib import Path
from contextlib import contextmanager
from typing import Callable, List, Optional, IO, Generator
from numpy import ndarray
from .autorg import (
    RG_RESULT,
    InsufficientDataError,
    NoGuinierRegionError,
)
from .sasio import (
    load_scattering_data,
    convert_inverse_angstrom_to_nanometer,
)
from .sas_argparser import GuinierParser, SASParser


def set_logging_level(verbose_flag: int) -> None:
    """
    Set logging level according to verbose flag of argparser
    :param verbose_flag: int flag for logging level
    """
    if verbose_flag == 1:
        logging.root.setLevel(logging.INFO)
    elif verbose_flag >= 2:
        logging.root.setLevel(logging.DEBUG)


def collect_files(file_list: List[str]) -> List[Path]:
    """
    Take file list from argparser and return list of paths
    :param file_list: file list as returned by the argparser
    :return: A list of Path objects which includes only existing files
    """
    files = [Path(i) for i in file_list if os.path.exists(i)]
    if platform.system() == "Windows" and files == []:
        files = list(Path.cwd().glob(file_list[0]))
        files.sort()
    return files


@contextmanager
def get_output_destination(
    output_path: Optional[Path] = None,
) -> Generator[IO[str], None, None]:
    """
    Return file or stdout object to write output to
    :param output_path: None if output to stdout, else Path to outputfile
    :return: opened file with write access or sys.stdout
    """
    # pylint: disable=R1705
    if output_path is not None:
        with open(output_path, "w") as destination:
            yield destination
    else:
        yield sys.stdout


def get_linesep(output_destination: IO[str]) -> str:
    """
    Get the appropriate linesep depending on the output destination.
    :param output_destination: an IO object, e.g. an open file or stdout
    :return: string with the correct linesep
    """
    # pylint: disable=R1705
    if output_destination == sys.stdout:
        return os.linesep
    else:
        return "\n"


def get_guinier_header(
    linesep: str, output_format: Optional[str] = None
) -> str:
    """Return appropriate header line for selected output format
    :param output_format: output format from string parser
    :param linesep: correct linesep for chosen destination
    :return: a one-line string"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                (
                    "File",
                    "Rg",
                    "Rg StDev",
                    "I(0)",
                    "I(0) StDev",
                    "First point",
                    "Last point",
                    "Quality,Aggregated",
                )
            )
            + linesep
        )
    else:
        return ""

def get_dnn_header(
    linesep: str, output_format: Optional[str] = None
) -> str:
    """Return appropriate header line for selected output format
    :param output_format: output format from string parser
    :param linesep: correct linesep for chosen destination
    :return: a one-line string"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                (
                    "File",
                    "Rg",
                    "Dmax",
                )
            )
            + linesep
        )
    else:
        return ""


def rg_result_to_output_line(
    rg_result: RG_RESULT,
    afile: Path,
    linesep: str,
    output_format: Optional[str] = None,
) -> str:
    """Return result line formatted according to selected output format
    :param rg_result: Result of an rg fit
    :param afile: The name of the file that was processed
    :param output_format: The chosen output format
    :param linesep: correct linesep for chosen destination
    :return: a one-line string including linesep"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                [
                    f"{afile}",
                    f"{rg_result.Rg:6.4f}",
                    f"{rg_result.sigma_Rg:6.4f}",
                    f"{rg_result.I0:6.4f}",
                    f"{rg_result.sigma_I0:6.4f}",
                    f"{rg_result.start_point:3}",
                    f"{rg_result.end_point:3}",
                    f"{rg_result.quality:6.4f}",
                    f"{rg_result.aggregated:6.4f}",
                ]
            )
            + linesep
        )
    elif output_format == "ssv":
        return (
            " ".join(
                [
                    f"{rg_result.Rg:6.4f}",
                    f"{rg_result.sigma_Rg:6.4f}",
                    f"{rg_result.I0:6.4f}",
                    f"{rg_result.sigma_I0:6.4f}",
                    f"{rg_result.start_point:3}",
                    f"{rg_result.end_point:3}",
                    f"{rg_result.quality:6.4f}",
                    f"{rg_result.aggregated:6.4f}",
                    f"{afile}",
                ]
            )
            + linesep
        )
    else:
        return f"{afile} {rg_result}{linesep}"

def dnn_result_to_output_line(
    dnn_result: tuple,
    afile: Path,
    linesep: str,
    output_format: Optional[str] = None,
) -> str:
    """Return result line formatted according to selected output format
    :param dnn_result: Result of an dnn inference, 2 tuple
    :param afile: The name of the file that was processed
    :param output_format: The chosen output format
    :param linesep: correct linesep for chosen destination
    :return: a one-line string including linesep"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                [
                    f"{afile}",
                    f"{dnn_result[0]:6.4f}",
                    f"{dnn_result[1]:6.4f}",
                ]
            )
            + linesep
        )
    elif output_format == "ssv":
        return (
            " ".join(
                [
                    f"{dnn_result[0]:6.4f}",
                    f"{dnn_result[1]:6.4f}",
                    f"{afile}",
                ]
            )
            + linesep
        )
    else:
        return f"{afile} {dnn_result[0]} {dnn_result[1]}{linesep}"


def run_guinier_fit(
    fit_function: Callable[[ndarray], RG_RESULT],
    parser: GuinierParser,
    logger: logging.Logger,
) -> None:
    """
    reads in the data, performs the guinier fit with a given algotithm and
    creates the
    :param fit_function : A Guinier fit function data -> RG_RESULT
    :param parser: a function that returns the output of argparse.parse()
    :param logger: a Logger
    """
    args = parser.parse_args()
    set_logging_level(args.verbose)
    files = collect_files(args.file)
    logger.debug("%s input files", len(files))

    with get_output_destination(args.output) as output_destination:
        linesep = get_linesep(output_destination)

        output_destination.write(
            get_guinier_header(
                linesep,
                args.format,
            )
        )

        for afile in files:
            logger.info("Processing %s", afile)
            try:
                data = load_scattering_data(afile)
            except OSError:
                logger.error("Unable to read file %s", afile)
            except ValueError:
                logger.error("Unable to parse file %s", afile)
            else:
                if args.unit == "Å":
                    data = convert_inverse_angstrom_to_nanometer(data)
                try:
                    rg_result = fit_function(data)
                except (
                    InsufficientDataError,
                    NoGuinierRegionError,
                    ValueError,
                    IndexError,
                ) as err:
                    sys.stderr.write(
                        f"{afile}, {err.__class__.__name__}: {err} {os.linesep}"
                    )
                else:
                    res = rg_result_to_output_line(
                        rg_result,
                        afile,
                        linesep,
                        args.format,
                    )
                    output_destination.write(res)
                    output_destination.flush()
def run_dnn(
    parser: SASParser,
    logger: logging.Logger,
) -> None:
    """
    reads in the data, infer the DNN and creates the result
    :param parser: a function that returns the output of argparse.parse()
    :param logger: a Logger
    """
    from .dnn import Rg_Dmax # heavy import

    args = parser.parse_args()
    set_logging_level(args.verbose)
    files = collect_files(args.file)
    logger.debug("%s input files", len(files))

    with get_output_destination(args.output) as output_destination:
        linesep = get_linesep(output_destination)

        output_destination.write(
            get_dnn_header(
                linesep,
                args.format,
            )
        )

        for afile in files:
            logger.info("Processing %s", afile)
            try:
                data = load_scattering_data(afile)
            except OSError:
                logger.error("Unable to read file %s", afile)
            except ValueError:
                logger.error("Unable to parse file %s", afile)
            else:
                if args.unit == "Å":
                    data = convert_inverse_angstrom_to_nanometer(data)
                q, I = data.T[:2]
                try:
                    dnn_result = Rg_Dmax(q, I)
                except (
                    InsufficientDataError,
                    NoGuinierRegionError,
                    ValueError,
                    IndexError,
                ) as err:
                    sys.stderr.write(
                        f"{afile}, {err.__class__.__name__}: {err} {os.linesep}"
                    )
                else:
                    res = dnn_result_to_output_line(
                        dnn_result,
                        afile,
                        linesep,
                        args.format,
                    )
                    output_destination.write(res)
                    output_destination.flush()



import numpy as np
import json
import h5py
import zipfile
from .resources import resource_filename
import io
import os


# Activation functions
def tanh(x): 
    return np.tanh(x)

def relu(x): 
    return np.maximum(0, x)

def sigmoid(x): 
    return 1 / (1 + np.exp(-x))

def linear(x): 
    return x

# Mapping activation names to functions
activation_functions = {
    'tanh': tanh,
    'relu': relu,
    'sigmoid': sigmoid,
    'linear': linear
}

# Parse config.json
def parse_config(config_path):
    config = {}
    if isinstance(config_path, io.IOBase):
        config = json.load(config_path)
    elif os.path.exists(config_path):
        with open(config_path, 'r') as f:
            config = json.load(f)
    else:
        raise RuntimeError(f"config_path type {type(config_path)} not handled, got {config_path}")
    layer_dims = []
    activations = []
    
    for layer in config.get('config',{}).get('layers', {}):
        if layer['class_name'] == 'InputLayer':
            layer_dims.append(layer['config']['batch_shape'][1])
        elif layer['class_name'] == 'Dense':
            layer_dims.append(layer['config']['units'])
            activation = layer['config']['activation']
            activations.append(activation_functions[activation])
    
    return layer_dims, activations

# Load weights from model.weights.h5
def load_weights(weights_path, layer_dims):
    with h5py.File(weights_path, 'r') as f:
        params = []
        for i in range(len(layer_dims) - 1):
            layer_name = f'dense_{i}' if i > 0 else 'dense'
            W = f[f'layers/{layer_name}/vars/0'][:]
            b = f[f'layers/{layer_name}/vars/1'][:]
            params.extend([W, b])
    return params

# Forward propagation
def forward_propagation(X, params, activations):
    A = X
    L = len(params) // 2
    for l in range(L):
        W, b = params[2*l], params[2*l+1]
        Z = np.dot(A, W) + b
        A = activations[l](Z)
    return A


class DenseLayer:
    def __init__(self, weights, bias, activation):
        """ Constructor
        
        :param weight: 2d array of size input_size x output_size
        :param bias: 1d array of size output_size
        :param activation: name of the activation function"""
        assert weights.ndim == 2 
        self.weights = weights
        assert bias.ndim == 1
        assert weights.shape[1] == bias.size
        self.bias = bias
        if callable(activation):
            self.activation = activation
        else:
            assert activation in activation_functions
            self.activation = activation_functions[activation]

    @property
    def input_size(self):
        return self.weights.shape[0]

    @property
    def output_size(self):
        return self.weights.shape[1]

    def __repr__(self):
        return f"Dense layer {self.input_size} ==> {self.output_size}, activation {self.activation}"

    def forward(self, ary):
        """Forward Propagate the array and return the result of activation(Ax+b)"""
        Z = np.dot(ary, self.weights) + self.bias
        return self.activation(Z)
    __call__ = forward

class DNN:
    def __init__(self, *args):
        """Constructor
        
        :param args: list of dense layers
        """
        self.list = args
    def infer(self,ary):
        """Infer the neural network"""
        tmp = ary
        for layer in self.list:
            tmp = layer(tmp)
        return tmp
    __call__ = infer

def preprocess(q, I):
    """
    Preprocess the input data and infer Rg and Dmax using the DNN.
    
    :param dnn: An instance of the DNN class
    :param q: 1D array of q values
    :param I: 1D array of intensity values
    :param dI: 1D array of the intensity error values
    :param q_interp: 1D array of the interpolated q values
    :return: Rg and Dmax
    """
    
    # Define q_interp as a regularly spaced array in the range 0-4 nm^-1 with 1024 points
    q_interp = np.linspace(0, 4, 1024)

    # Normalize I by Imax
    Imax = I.max()
    I_normalized = I / Imax

    # Interpolate I over q_interp
    I_interp = np.interp(q_interp, q, I_normalized, left=1, right=0)
    return I_interp


def parse_keras_file(keras_file):
    with zipfile.ZipFile(keras_file, 'r') as z:
        with z.open('config.json') as config_file:
            #config = parse_config(io.TextIOWrapper(config_file, 'tf-8'))
            config = parse_config(config_file)
        with z.open('model.weights.h5') as weights_file:
            weights = load_weights(io.BytesIO(weights_file.read()), config[0])
            #weights = load_weights(weights_file, config[0])
    return config, weights


class KerasDNN:
    def __init__(self, keras_file):
        config, weights = parse_keras_file(keras_file)
        self.dnn =  DNN(*[DenseLayer(weights[2*i], weights[2*i+1], a) for i,a in enumerate(config[1])])
                
    def infer(self, q, I):
        """Infer the neural network with q/I
        :param q: 1D array, in inverse nm
        :param I: 1D array, same size as q
        :return: result of the neural network.
        """
        Iprep = preprocess(q,I)
        output = self.dnn.infer(Iprep)
        # Extract Rg and Dmax from the output
        Rg = output[0]  # Assuming Rg is the first element
        Dmax = output[1]  # Assuming Dmax is the second element
    
        return Rg, Dmax
    __call__ = infer

Rg_Dmax = KerasDNN(resource_filename("keras_models/Rg+Dmax.keras"))


#!usr/bin/env python
# coding: utf-8

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__date__ = "28/11/2023"
__copyright__ = "2015-2021, ESRF"

import os
import logging
logger = logging.getLogger("utilstest")
from silx.resources import ExternalResources
downloader = ExternalResources("freesas", "http://www.silx.org/pub/freesas/testdata", "FREESAS_TESTDATA")


def get_datafile(name):
    """Provides the full path of a test file,
    downloading it from the internet if needed

    :param name: name of the file to get
    :return: full path of the datafile
    """
    logger.info(f"Download file {name}")
    fullpath = downloader.getfile(name)
    return fullpath

def clean():
    pass

class TestOptions(object):
    """
    Class providing useful stuff for preparing tests.
    """
    def __init__(self):
        self.TEST_LOW_MEM = False
        """Skip tests using too much memory"""

        self.TEST_RANDOM = False
        """Use a random seed to generate random values"""

        self.options = None

    def __repr__(self):
        return f"TEST_LOW_MEM={self.TEST_LOW_MEM} "\
               f"TEST_RANDOM={self.TEST_RANDOM} "
    @property
    def low_mem(self):
        """For compatibility"""
        return self.TEST_LOW_MEM

    def configure(self, parsed_options=None):
        """Configure the TestOptions class from the command line arguments and the
        environment variables
        """

        if parsed_options is not None and parsed_options.low_mem:
            self.TEST_LOW_MEM = True
        elif os.environ.get('FREESAS_LOW_MEM', 'True') == 'False':
            self.TEST_LOW_MEM = True

        if parsed_options is not None and parsed_options.random:
            self.TEST_RANDOM = True
        if os.environ.get('FREESAS_RANDOM', 'False').lower() in ("1", "true", "on"):
            self.TEST_RANDOM = True


    def add_parser_argument(self, parser):
        """Add extrat arguments to the test argument parser

        :param ArgumentParser parser: An argument parser
        """
        parser.add_argument("-l", "--low-mem", dest="low_mem", default=False,
                            action="store_true",
                            help="Disable test with large memory consumption (>100Mbyte")
        parser.add_argument("-r", "--random", dest="random", default=False,
                            action="store_true",
                            help="Enable actual random number to be generated. By default, stable seed ensures reproducibility of tests")


test_options = TestOptions() #singleton



"""End to end tests for auto_gpa.py """

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "29/11/2023"

import unittest
import pathlib
import logging
from platform import system
from subprocess import run, PIPE, STDOUT
from os import linesep
from os.path import normpath
from freesas.test.utilstest import get_datafile

logger = logging.getLogger(__name__)

if system() == "Windows":
    cormapy = "cormapy.exe"
else:
    cormapy = "cormapy"


class TestCormap(unittest.TestCase):
    """End to end tests for free_bift"""

    cwd = pathlib.Path.cwd()
    bsa_filename = pathlib.Path(get_datafile("bsa_005_sub.dat"))

    def test_1_input_file_results_no_correlation(self):
        """
        Test that cormap does not list a correlation if only one file is provided.
        """

        run_app = run(
            [cormapy, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        if system() == "Windows":
            run_app_output_raw = str(run_app.stdout)[:-1].replace("\\\\", "\\")
            run_app_output = [
                line.replace("\\n", "").replace("\\r", "")
                for line in run_app_output_raw.split("\\n")[:-1]
            ]
        else:
            run_app_output_raw = str(run_app.stdout, "utf-8")[:-1]
            run_app_output = run_app_output_raw.split(linesep)[:-1]

        self.assertEqual(
            run_app.returncode, 0, msg="cormapy on BM29 BSA completed well"
        )

        corr_table_head_pos = [
            line_number
            for line_number, line in enumerate(run_app_output)
            if "Pr(>C)" in line
        ][0]

        self.assertEqual(
            run_app_output[corr_table_head_pos + 1],
            "",
            msg="Correlation table is empty",
        )

    def test_correlation_of_file_with_itself_is_1(self):
        """
        Test that the auto-correlation of a file is 1.
        """

        run_app = run(
            [
                cormapy,
                normpath(str(self.bsa_filename)),
                normpath(str(self.bsa_filename)),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )

        if system() == "Windows":
            run_app_output_raw = str(run_app.stdout)[:-1].replace("\\\\", "\\")
            run_app_output = [
                line.replace("\\n", "").replace("\\r", "")
                for line in run_app_output_raw.split("\\n")[:-1]
            ]
        else:
            run_app_output_raw = str(run_app.stdout, "utf-8")[:-1]
            run_app_output = run_app_output_raw.split(linesep)[:-1]

        self.assertEqual(
            run_app.returncode,
            0,
            msg="auto-correlation cormapy on BM29 BSA completed well",
        )

        corr_table_head_pos = [
            line_number
            for line_number, line in enumerate(run_app_output)
            if "Pr(>C)" in line
        ][0]

        self.assertEqual(
            run_app_output[corr_table_head_pos + 1].split(),
            [
                "1",
                "vs.",
                "2",
                "0",
                "1.000000",
            ],
            msg="Auto-correaltion is 1",
        )


def suite():
    """Build test suite for free_bift"""
    test_suite = unittest.TestSuite()

    test_suite.addTest(TestCormap("test_1_input_file_results_no_correlation"))
    test_suite.addTest(TestCormap("test_correlation_of_file_with_itself_is_1"))

    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""This replaces the `cimport numpy` for C-level datatypes"""

from libc.stdint cimport int8_t, uint8_t, int16_t, uint16_t, \
                         int32_t, uint32_t, int64_t, uint64_t

ctypedef double float64_t
ctypedef float float32_t

from cython cimport floating

ctypedef fused any_int_t:
    uint8_t
    uint16_t
    uint32_t
    uint64_t
    int8_t
    int16_t
    int32_t
    int64_t

ctypedef fused any_t:
    int
    long
    uint8_t
    uint16_t
    uint32_t
    uint64_t
    int8_t
    int16_t
    int32_t
    int64_t
    float32_t
    float64_t



#Cython module to calculate distances of set of atoms

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2023, ESRF"

import sys
import cython
import numpy
from cython cimport floating
from libc.math cimport sqrt, fabs, exp


@cython.wraparound(False)
@cython.boundscheck(False)
def calc_invariants(floating[:, :] atoms):
    """
    Calculate the invariants of the structure, i.e fineness, radius of gyration and diameter of the model.

    Nota: to economize size*numpy.sqrt, the sqrt is taken at the end of the calculation. 
    We should have done s += sqrt(d) and then s/size, but we do s+= d and then sqrt(s/size).
    You can check that the result is the same.

    @param atoms: 2d-array with atom coordinates:[[x,y,z],...]
    @return: 3-tuple containing (fineness, Rg, Dmax)
        * average distance between an atoms and its nearest neighbor
        * radius of gyration of the model
        * diameter of the model
    """
    cdef:
        int i, j, size
        floating d, x1, y1, z1, dx, dy, dz, big, d2, sum_d2, d2max
    size = atoms.shape[0]
    assert atoms.shape[1] >= 3
    big = sys.maxsize
    s = 0.0
    sum_d2 = 0.0 
    d2max = 0.0
    for i in range(size):
        x1 = atoms[i, 0]
        y1 = atoms[i, 1]
        z1 = atoms[i, 2]
        d = big
        for j in range(size):
            if i == j:
                continue
            dx = atoms[j, 0] - x1
            dy = atoms[j, 1] - y1
            dz = atoms[j, 2] - z1
            d2 = dx * dx + dy * dy + dz * dz
            sum_d2 += d2
            d2max = max(d2max, d2)
            d = min(d, d2)
        s += d
    return sqrt(s / size), sqrt(sum_d2 / 2.0) / size, sqrt(d2max)


cdef inline floating hard_sphere(floating pos, floating radius) noexcept nogil:
    """Density using hard spheres
    @param pos: fabs(d1-d)
    """
    if pos > 2.0 * radius:
        return 0.0 
    return (4 * radius + pos) * (2 * radius - pos) ** 2 / (16.0 * radius ** 3)

cdef inline floating soft_sphere(floating pos, floating radius) noexcept nogil:
    """Density using soft spheres (gaussian density)
    @param pos: fabs(d1-d)
    @param radius: radius of the equivalent hard sphere
    """
    cdef floating sigma = 0.40567 * radius
    return exp(- pos * pos / (2.0 * sigma * sigma)) * 0.3989422804014327 / sigma


@cython.wraparound(False)
@cython.boundscheck(False)
def calc_distance(floating[:,:] atoms1, floating[:,:] atoms2, floating fineness1, floating fineness2):
    """
    Calculate the Normalized Spatial Discrepancy (NSD) between two molecules
    
    @param atoms1,atoms2: 2d-array with atom coordinates[[x,y,z],...]
    @param fineness1, fineness2: fineness of each molecule
    @return: NSD atoms1-atoms2
    """
    cdef:
        int i, j, size1=atoms1.shape[0], size2=atoms2.shape[0]
        double d, d2, dx, dy, dz, x1, y1, z1
        double s1=0.0, s2=0.0, big = sys.maxsize
        double[:] min_col = numpy.zeros(size2, numpy.float64) + big
    assert atoms1.shape[1] >= 3
    assert atoms2.shape[1] >= 3
    
    for i in range(size1):
        x1 = atoms1[i,0]
        y1 = atoms1[i,1]
        z1 = atoms1[i,2]
        d = big
        
        for j in range(size2):            
            dx = atoms2[j,0] - x1
            dy = atoms2[j,1] - y1
            dz = atoms2[j,2] - z1
            d2 = dx*dx + dy*dy + dz*dz
            d = min(d, d2)
            min_col[j] = min(min_col[j], d2) 
        s1 += d
    for j in range(size2):
        s2 += min_col[j]
    
    return sqrt(0.5*((1./(size1*fineness2*fineness2))*s1 + (1./(size2*fineness1*fineness1))*s2))

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
def calc_density(floating[:, :] atoms, floating dmax, int npt, floating r=0.0, bint hard=True):
    """
    Calculate the density rho(r)
    
    #TODO: formula for rigid sphere:
    A = (4*R+d)*(2*R-d)**2/16.0/R**3
    
    @param atoms: 2d-array with atom coordinates[[x,y,z],...]
    @param dmax: Diameter of the model
    @param npt: number of point in the density
    @param r: radius of an atom
    @param hard: use hard spheres model
    @return: 1d-array of 
    """
    
    cdef:
        int i, j, k, size = atoms.shape[0]        
        int width = 1 if hard else 2 
        floating d, dmax_plus, dx, dy, dz, x1, y1, z1
        floating delta, d_min, d_max, d1, den 
        double[::1] out = numpy.zeros(npt, numpy.float64)
        double s
        
    assert atoms.shape[1] >= 3
    assert size > 0
    assert dmax > 0
    dmax_plus = dmax * (1.0 + numpy.finfo(numpy.float32).eps)
    delta = dmax_plus / npt
    
    for i in range(size, nogil=True):
        x1 = atoms[i, 0]
        y1 = atoms[i, 1]
        z1 = atoms[i, 2]
        for j in range(size):
            dx = atoms[j, 0] - x1
            dy = atoms[j, 1] - y1
            dz = atoms[j, 2] - z1
            d = sqrt(dx * dx + dy * dy + dz * dz)
            d_min = max(0.0, d - width * r)
            d_max = min(dmax, d + width * r)
            for k in range(<int>(d_min / delta), <int>(d_max / delta)+1):
                if hard:
                    out[k] += hard_sphere(fabs(k * delta - d), r)
                else:
                    out[k] += soft_sphere(fabs(k * delta - d), r)
    return numpy.asarray(out)




#!/usr/bin/python3
# coding: utf-8
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = ["Jérôme Kieffer", "Mayank Yadav"]
__license__ = "MIT"
__copyright__ = "2024, ESRF"
__date__ = "11/09/2024"

import sys
import logging
from freesas.sas_argparser import SASParser
from freesas.fitting import run_dnn

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("free_dnn")

if sys.version_info < (3, 6):
    logger.error("This code uses F-strings and requires Python 3.6+")


def build_parser() -> SASParser:
    """Build parser for input and return list of files.
    :return: parser
    """
    description = (
        "Assess the radius of gyration (Rg) and the diameter of the particle (Dmax) using a Dense Neural-Network"
        " for a set of scattering curves"
    )
    epilog = """free_dnn is an alternative implementation of
    `gnnom` (https://doi.org/10.1016/j.str.2022.03.011).
    As this tool used a different training set, some results are likely to differ.
    """
    parser = SASParser(prog="free_gpa", description=description, epilog=epilog)
    file_help_text = "dat files of the scattering curves"
    parser.add_file_argument(help_text=file_help_text)
    parser.add_output_filename_argument()
    parser.add_output_data_format("native", "csv", "ssf", default="native")
    parser.add_q_unit_argument()

    return parser



def main() -> None:
    """Entry point for free_gpa app"""
    parser = build_parser()
    run_dnn(parser=parser, logger=logger)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2024-2024  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["Jérôme Kieffer"]
__license__ = "MIT"
__date__ = "28/06/2024"

import unittest
import logging
import os
from ..resources import resource_filename
logger = logging.getLogger(__name__)

class TestResources(unittest.TestCase):

    def test_filename(self):
        """
        Test for returning the actual path of an existing model 
        """
        self.assertTrue(os.path.exists(resource_filename("keras_models/Rg+Dmax.keras")),
                        msg="file exists")


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(TestResources("test_filename"))
    return test_suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2024-2024  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["Jérôme Kieffer"]
__license__ = "MIT"
__date__ = "03/07/2024"

import unittest
import logging
import os
import io
import numpy as np
from .utilstest import get_datafile
from ..resources import resource_filename
from ..sasio import load_scattering_data
from ..dnn import *

logger = logging.getLogger(__name__)

class TestDNN(unittest.TestCase):

    def test_activation_functions(self):
        """
        Test for the activation functions
        """
        x = np.array([-1, 0, 1])
        
        # Test tanh
        expected_tanh = np.tanh(x)
        self.assertTrue(np.allclose(tanh(x), expected_tanh), msg="tanh function failed")
        
        # Test relu
        expected_relu = np.maximum(0, x)
        self.assertTrue(np.allclose(relu(x), expected_relu), msg="relu function failed")
        
        # Test sigmoid
        expected_sigmoid = 1 / (1 + np.exp(-x))
        self.assertTrue(np.allclose(sigmoid(x), expected_sigmoid), msg="sigmoid function failed")
        
        # Test linear
        expected_linear = x
        self.assertTrue(np.allclose(linear(x), expected_linear), msg="linear function failed")
        
        logger.info("test_activation_functions ran successfully")

    
    def test_preprocess(self):
        """
        Test for the preprocessing function 
        """
        datfile = get_datafile("bsa_005_sub.dat")
        data = load_scattering_data(datfile)
        q, I, sigma = data.T
        Iprep = preprocess(q, I)
        self.assertEqual(Iprep.max(), 1, msg="range 0-1")
        self.assertEqual(Iprep.shape, (1024,), msg="size 1024")
 

    
    def test_forward_propagation(self):
        """
        Test for the forward_propagation function
        """
        try:
            X = np.random.rand(1, 10)
            params = [np.random.rand(10, 20), np.random.rand(20), np.random.rand(20, 10), np.random.rand(10)]
            activations = [np.tanh, np.tanh]
            output = forward_propagation(X, params, activations)
            self.assertEqual(output.shape, (1, 10))
            logger.info("test_forward_propogation ran successfully")
        except Exception as e:
            logger.error(f"test_forward_propagation failed: {e}")
            raise

    
    def test_DenseLayer(self):
        """
        Test for the DenseLayer class
        """
        try :
            weights = np.random.rand(10, 20)
            bias = np.random.rand(20)
            layer = DenseLayer(weights, bias, 'tanh')
            self.assertEqual(layer.input_size, 10)
            self.assertEqual(layer.output_size, 20)
            output = layer.forward(np.random.rand(1, 10))
            self.assertEqual(output.shape, (1, 20))
            logger.info("test_DenseLayer ran successfully")
        except Exception as e:
            logger.error(f"test_DenseLayer failed: {e}")
            raise

    
    
    def test_DNN(self):
        """
        Test for the DNN class
        """
        try:
            layers = [DenseLayer(np.random.rand(10, 20), np.random.rand(20), 'tanh'),
                      DenseLayer(np.random.rand(20, 10), np.random.rand(10), 'tanh')]
            dnn = DNN(*layers)
            output = dnn.infer(np.random.rand(1, 10))
            self.assertEqual(output.shape, (1, 10))
            logger.info("test_DNN ran successfully")
        except Exception as e:
            logger.error(f"test_DNN failed: {e}")
            raise


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    test_suite = unittest.TestSuite()
    test_suite.addTest(loader(TestDNN))
    # test_suite.addTest(TestDNN("test_preprocess"))
    return test_suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())
